/***********************************************************************
*
* mimedefang.c
*
* C interface to the attachment-filter program for stripping or altering
* MIME attachments in incoming Sendmail connections.
*
* Copyright (C) 2000 Roaring Penguin Software Inc.
* http://www.roaringpenguin.com
*
* This program may be distributed under the terms of the GNU General
* Public License, Version 2, or (at your option) any later version.
*
* This program was derived from the sample mail filter included in
* libmilter/README in the Sendmail 8.11 distribution.
***********************************************************************/

static char const RCSID[] =
"$Id: mimedefang.c,v 1.35 2001/08/02 14:52:30 dfs Exp $";

#define _BSD_SOURCE 1

#ifdef __linux__
/* On Linux, we need this defined to get fdopen.  On BSD, if we define
 * it, we don't get all the u_char, u_long, etc definitions.  GRR! */
#define _POSIX_SOURCE 1
#endif

#include "config.h"

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>

#include "libmilter/mfapi.h"

#include <stdio.h>

#ifdef ENABLE_DEBUGGING
#include <signal.h>
#define DEBUG(x) x
#else
#define DEBUG(x) (void) 0
#endif

#define DEBUG_ENTER(func, line) DEBUG(syslog(LOG_INFO, "Entering %s (line %d)", func, line))
#define DEBUG_EXIT(func, line, ret) DEBUG(syslog(LOG_INFO, "Exiting %s (line %d) ret=%s", func, line, ret))

/* Keep track of private data -- file name and fp for writing e-mail body */

int DebugMode = 0;

struct privdata {
    char *hostname;		/* Name of connecting host */
    char *hostip;		/* IP address of connecting host */
    char *dir;			/* Work directory */
    FILE *fp;			/* File for message body */
    FILE *headerFP;		/* File for message headers */
    FILE *recipFP;		/* File for recipients */
};

static sfsistat cleanup(SMFICTX *ctx);
static int IncrementRunningFilters(void);
static int DecrementRunningFilters(void);
static sfsistat mfclose(SMFICTX *ctx);

int DirCount = 0;
#define DATA ((struct privdata *) smfi_getpriv(ctx))

/* How many tries to get a unique directory name before giving up? */
#define MAXTRIES 8192

/* Size of buffer for temporary string manipulation */
#define SMALLBUF 256

/* Size of chunk when replacing body */
#define CHUNK 4096

/* Number of file descriptors to close when forking */
#define CLOSEFDS 256

/* Number of running Perl filters */
static volatile int NumRunningFilters = 0;

/* Maximum allowable number of filters (essentially infinite) */
static int MaxRunningFilters = 1000000;

/* Mutex to protect access to NumRunningFilters */
static pthread_mutex_t FilterMutex = PTHREAD_MUTEX_INITIALIZER;

#ifdef ENABLE_DEBUGGING
/**********************************************************************
*%FUNCTION: handle_sig
*%ARGUMENTS:
* s -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Handler for SIGSEGV and SIGBUS -- logs a message and returns -- hopefully,
* we'll get a nice core dump the second time around
***********************************************************************/
static void
handle_sig(int s)
{
    syslog(LOG_ERR, "WHOA, NELLY!  Caught signal %d -- this is bad news.  Core dump at 11.", s);

    /* Default is terminate and core. */
    signal(s, SIG_DFL);

    /* Return and probably cause core dump */
}
#endif

/**********************************************************************
*%FUNCTION: closefiles
*%ARGUMENTS:
* None
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Hack -- closes a whole bunch of file descriptors.  Use after a fork()
*         to conserve descriptors.
***********************************************************************/
static void
closefiles(void)
{
    int i;
    for (i=0; i<CLOSEFDS; i++) {
	(void) close(i);
    }
}

/**********************************************************************
*%FUNCTION: chomp
*%ARGUMENTS:
* str -- a string
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Removes a trailing newline (if any) from str
***********************************************************************/
void
chomp(char *str)
{
    size_t len = strlen(str);
    if (len && (str[len-1] == '\n')) {
	str[len-1] = 0;
    }
}

/**********************************************************************
*%FUNCTION: IncrementRunningFilters
*%ARGUMENTS:
* None
*%RETURNS:
* -1 if we've hit maximum number of filters, 0 otherwise.
*%DESCRIPTION:
* Increments number of running filters using a mutex to lock access.
***********************************************************************/
static int
IncrementRunningFilters(void)
{
    if (pthread_mutex_lock(&FilterMutex)) {
	DEBUG(syslog(LOG_INFO, "IncrementRunningFilters: error locking FilterMutex"));
	return -1;
    }
    if (NumRunningFilters >= MaxRunningFilters) {
	syslog(LOG_INFO, 
	       "Maximum number of filters (%d) running", MaxRunningFilters);
	(void) pthread_mutex_unlock(&FilterMutex);
	return -1;
    }
    NumRunningFilters++;
    DEBUG(syslog(LOG_INFO, "New filter; now running %d filters", NumRunningFilters));
    (void) pthread_mutex_unlock(&FilterMutex);
    return 0;
}

/**********************************************************************
*%FUNCTION: DecrementRunningFilters
*%ARGUMENTS:
* None
*%RETURNS:
* -1 if there was an error; 0 otherwise.
*%DESCRIPTION:
* Decrements number of running filters using a mutex to lock access.
***********************************************************************/
static int
DecrementRunningFilters(void)
{
    if (pthread_mutex_lock(&FilterMutex)) {
	DEBUG(syslog(LOG_INFO, "DecrementRunningFilters: error locking FilterMutex"));
	return -1;
    }
    if (NumRunningFilters == 0) {
	/* Shouldn't happen! */
	DEBUG(syslog(LOG_ERR, "DecrementRunningFilters: NumRunningFilters is zero!"));
	(void) pthread_mutex_unlock(&FilterMutex);
	return -1;
    }
    NumRunningFilters--;
    DEBUG(syslog(LOG_INFO, "Filter finished; now running %d filters", NumRunningFilters));
    (void) pthread_mutex_unlock(&FilterMutex);
    return 0;
}

/**********************************************************************
*%FUNCTION: mfconnect
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* hostname -- name of connecting host
* sa -- socket address of connecting host
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Allocates a private data structure for tracking this connection
***********************************************************************/
static sfsistat
mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
{
    struct privdata *data;
    struct sockaddr_in *insa = (struct sockaddr_in *) sa;
    char *tmp;

    DEBUG_ENTER("mfconnect", __LINE__);
    /* Delete any existing context data */
    mfclose(ctx);
    
    /* If too many running filters, reject connection at this phase */
    if (NumRunningFilters >= MaxRunningFilters) {
	syslog(LOG_INFO, "mfconnect: Maximum number of filters (%d) running",
	       MaxRunningFilters);
	DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    data = malloc(sizeof *data);
    if (!data) {
	DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->hostname = NULL;
    data->hostip   = NULL;
    data->dir      = NULL;
    data->fp       = NULL;
    data->headerFP = NULL;
    data->recipFP  = NULL;

    /* Save private data */
    smfi_setpriv(ctx, data);
    if (hostname) {
	data->hostname = strdup(hostname);
	if (!data->hostname) {
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    } else {
	data->hostname = NULL;
    }

    /* Padding -- should be big enough for IPv6 addresses, but currently
       only support IPv4 -- see inet_ntoa below.*/
    if (!sa) {
	data->hostip = NULL;
    } else {
	data->hostip = malloc(65);
	if (!data->hostip) {
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	tmp = inet_ntoa(insa->sin_addr);
	if (!tmp) {
	    mfclose(ctx);
	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	strncpy(data->hostip, tmp, 64);
	data->hostip[64] = 0;
    }
	
    data->dir = NULL;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;

}

/**********************************************************************
*%FUNCTION: envfrom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* from -- list of arguments to "MAIL FROM:" SMTP command.
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Allocates a private data structure for tracking this message, and
* opens a temporary file for saving message body.
***********************************************************************/
static sfsistat
envfrom(SMFICTX *ctx, char **from)
{
    struct privdata *data;
    int fd;
    int tries;
    int success;
    char buffer[SMALLBUF];
    char buf2[SMALLBUF];
    FILE *fp;

    DEBUG_ENTER("envfrom", __LINE__);
    /* Get the private context */
    data = DATA;
    if (!data) {
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* We need a nice clean directory all our own.  It may seem that
     * using DirCount in this way is dangerous because of thread-safety,
     * but it doesn't really matter.  We just need something that
     * increments, and if we miss a number or two, who cares?  If two
     * threads get the same number, the pthread_self() will disambiguate
     * them. */

    success = 0;
    for (tries = 0; tries < MAXTRIES; tries++) {
	++DirCount;
	snprintf(buffer, SMALLBUF, "%s/mdefang-%ld-%d",
		 SPOOLDIR,
		 (long) pthread_self(), DirCount);
	if (!mkdir(buffer, 0700)) {
	    success = 1;
	    break;
	}
	if (errno != EEXIST) {
	    break;
	}
    }

    /* Could not create temp. directory */
    if (!success) {
	mfclose(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    if (data->dir) free(data->dir);
    if (data->fp) fclose(data->fp);
    if (data->headerFP) fclose(data->headerFP);
    if (data->recipFP) fclose(data->recipFP);

    data->dir = NULL;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    data->dir = strdup(buffer);

    if (!data->dir) {
	/* Don't forget to clean up directory... */
	rmdir(buffer);
	mfclose(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Write host name and host IP */
    snprintf(buf2, SMALLBUF, "%s/HOSTNAME", buffer);
    fp = fopen(buf2, "w");
    if (fp) {
	if (data->hostname) {
	    fprintf(fp, "%s\n", data->hostname);
	} else {
	    fprintf(fp, "*UNKNOWN*\n");
	}
	fclose(fp);
    }
    snprintf(buf2, SMALLBUF, "%s/HOSTIP", buffer);
    fp = fopen(buf2, "w");
    if (fp) {
	if (data->hostip) {
	    fprintf(fp, "%s\n", data->hostip);
	} else {
	    fprintf(fp, "*UNKNOWN*\n");
	}
	fclose(fp);
    }

    /* Synthesize filename for output */
    snprintf(buf2, SMALLBUF, "%s/INPUTMSG", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->fp = fdopen(fd, "w+")) == NULL) {
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/HEADERS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->headerFP = fdopen(fd, "w+")) == NULL) {
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/RECIPIENTS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->recipFP = fdopen(fd, "w+")) == NULL) {
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Write the sender */
    snprintf(buf2, SMALLBUF, "%s/SENDER", data->dir);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(fp = fdopen(fd, "w+")) == NULL) {
	cleanup(ctx);
	DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    fprintf(fp, "%s\n", from[0]);
    fclose(fp);

    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: rcptto
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* to -- list of arguments to each RCPT_TO
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Saves recipient data
***********************************************************************/
static sfsistat
rcptto(SMFICTX *ctx, char **to)
{
    DEBUG_ENTER("rcptto", __LINE__);
    if (!DATA) {
	DEBUG_EXIT("rcptto", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    fprintf(DATA->recipFP, "%s\n", to[0]);
    DEBUG_EXIT("rcptto", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: header
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* headerf -- Header field name
* headerv -- Header value
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes the header to the temporary file
***********************************************************************/
static sfsistat
header(SMFICTX *ctx, char *headerf, char *headerv)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("header", __LINE__);
    if (!data) {
	DEBUG_EXIT("header", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* We can close recipFP now to save a descriptor */
    if (data->recipFP && (fclose(data->recipFP) == EOF)) {
	data->recipFP = NULL;
	DEBUG_EXIT("header", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->recipFP = NULL;

    /* Write the header to the message file */
    fprintf(data->fp, "%s: %s\r\n", headerf, headerv);

    /* Write to our HEADERS file, but chew up embedded carriage returns or
       newlines */
    fprintf(data->headerFP, "%s: ", headerf);
    while (*headerv) {
	if (*headerv != '\n' && *headerv != '\r') {
	    putc(*headerv, data->headerFP);
	}
	headerv++;
    }
    fprintf(data->headerFP, "\n");
    DEBUG_EXIT("header", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eoh
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a blank line to indicate the end of headers.
***********************************************************************/
static sfsistat
eoh(SMFICTX *ctx)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("eoh", __LINE__);
    if (!data) {
	DEBUG_EXIT("eoh", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* We can close headerFP to save a descriptor */
    if (data->headerFP && (fclose(data->headerFP) == EOF)) {
	data->headerFP = NULL;
	DEBUG_EXIT("eoh", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    data->headerFP = NULL;

    fprintf(data->fp, "\r\n");
    DEBUG_EXIT("eoh", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: body
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* text -- a chunk of text from the mail body
* len -- length of chunk
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a chunk of the body to the temporary file
***********************************************************************/
static sfsistat
body(SMFICTX *ctx, u_char *text, size_t len)
{
    DEBUG_ENTER("body", __LINE__);
    if (!DATA) {
	DEBUG_EXIT("body", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    if (fwrite(text, len, 1, DATA->fp) <= 0) {
	DEBUG(syslog(LOG_ERR, "fwrite failed: %m line %d", __LINE__));
	cleanup(ctx);
	DEBUG_EXIT("body", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    DEBUG_EXIT("body", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* This is where all the action happens.  Called at end of message, it
* forks the Perl scanner which may or may not ask for the body to be
* replaced.
***********************************************************************/
static sfsistat
eom(SMFICTX *ctx)
{
    pid_t n;
    int status;
    char buffer[SMALLBUF];
    FILE *fp;
    struct privdata *data = DATA;
    int r;
    int problem = 0;

    DEBUG_ENTER("eom", __LINE__);
    /* Close output file */
    if (!data) {
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    if (!data->fp || (fclose(data->fp) == EOF)) problem = 1;
    if (data->headerFP && (fclose(data->headerFP) == EOF)) problem = 1;
    if (data->recipFP && (fclose(data->recipFP) == EOF)) problem = 1;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    if (problem) {
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }

    /* Create a sanity-check file */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if ((fp = fopen(buffer, "w")) != NULL) {
	fclose(fp);
    }

    /* Run the Perl filter */
    if (IncrementRunningFilters() < 0) {
	/* Doh!  Too many running filters... Sorry, have to abandon at
	   this point */
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    n = fork();
    if (n < 0) {
	syslog(LOG_WARNING, "fork() failed -- cannot run filter");
	DecrementRunningFilters();
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    } else if (n == 0) {
	/* In the child */
	closefiles();
	execl(PERL_PATH, "perl", MIMEDEFANG_PL, data->dir, NULL);
	_exit(EXIT_FAILURE);
    } else {
	/* In the parent */
	if (waitpid(n, &status, 0) < 0) {
	    DecrementRunningFilters();
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	if (DecrementRunningFilters() < 0) {
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
    }

    /* If BOUNCE file is there, bounce the message */
    snprintf(buffer, SMALLBUF, "%s/BOUNCE", data->dir);
    if (!access(buffer, F_OK)) {
	int fd, n;
	fd = open(buffer, O_RDONLY);
	if (fd < 0) {
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	if ((n=read(fd, buffer, SMALLBUF-1)) < 0) {
	    close(fd);
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	close(fd);
	if (!n) ++n;
	buffer[n-1] = 0;
	smfi_setreply(ctx, "554", "5.7.1", buffer);
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_REJECT");
	return SMFIS_REJECT;
    }

    /* If DISCARD file is there, silently discard the message */
    snprintf(buffer, SMALLBUF, "%s/DISCARD", data->dir);
    if (!access(buffer, F_OK)) {
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_DISCARD");
	return SMFIS_DISCARD;
    }

    /* If TEMPFAIL file is there, tempfail the message */
    snprintf(buffer, SMALLBUF, "%s/TEMPFAIL", data->dir);
    if (!access(buffer, F_OK)) {
	cleanup(ctx);
	DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	return SMFIS_TEMPFAIL;
    }
    
    /* If body is changed, replace it */
    snprintf(buffer, SMALLBUF, "%s/CHANGED", data->dir);
    if (!access(buffer, F_OK)) {
	int fd;
	int j;
	unsigned char chunk[CHUNK];
	snprintf(buffer, SMALLBUF, "%s/NEWBODY", data->dir);
	fd = open(buffer, O_RDONLY);
	if (fd < 0) {
	    cleanup(ctx);
	    DEBUG_EXIT("eom", __LINE__, "SMFIS_TEMPFAIL");
	    return SMFIS_TEMPFAIL;
	}
	while ((j=read(fd, chunk, CHUNK)) > 0) {
	    smfi_replacebody(ctx, chunk, j);
	}
	close(fd);
    }

    /* Do we need to change the Content-Type header? WARNING: WE CAN
       ONLY CHANGE A SINGLE-PART MESSAGE TO MULTIPART, BUT THE C CODE DOES
       NOT ENFORCE THIS! */
    snprintf(buffer, SMALLBUF, "%s/CONTENT-TYPE", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
#if SMFI_VERSION == 1
	/* Cannot change headers in Sendmail 8.10 -- please upgrade
	   to 8.11 */
#warning Please upgrade to Sendmail 8.11 -- Sendmail 8.10 with MIMEDefang will not handle single-part messages correctly.
#else
	fgets(buffer, SMALLBUF, fp);
	chomp(buffer);
	if (strlen(buffer) > 0) {
	    smfi_chgheader(ctx, "Content-Type", 1, buffer);
	}
	/* We've just changed the message to multi-part.  Better delete 
	   other MIME headers which may be present and confuse mail clients */
	smfi_chgheader(ctx, "Content-Disposition", 1, "");
	smfi_chgheader(ctx, "Content-Transfer-Encoding", 1, "");
#endif
	fclose(fp);
    }
    /* If there are new headers, add them */
    snprintf(buffer, SMALLBUF, "%s/NEWHEADERS", data->dir);
    fp = fopen(buffer, "r");
    if (fp) {
	char header[SMALLBUF];
	char value[SMALLBUF];
	while (fgets(header, SMALLBUF, fp) && fgets(value, SMALLBUF, fp)) {
	    chomp(header);
	    chomp(value);
	    smfi_addheader(ctx, header, value);
	}
	fclose(fp);
    }

    /* If UNFINISHED file is still there, ooops... */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if (!access(buffer, F_OK)) {
	DEBUG(syslog(LOG_ERR, "Perl process did not remove UNFINISHED file"));
	smfi_addheader(ctx, "X-MIMEDefang-Warning", "Perl process did not finish");
    }
    smfi_addheader(ctx, "X-Scanned-By", "MIMEDefang " VERSION " (www dot roaringpenguin dot com slash mimedefang)");
    r = cleanup(ctx);
    DEBUG(syslog(LOG_INFO, "Exiting eom (line %d) ret=%d", __LINE__, r));
    return r;
}

/**********************************************************************
*%FUNCTION: mfclose
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_ACCEPT
*%DESCRIPTION:
* Called when connection is closed.
***********************************************************************/
static sfsistat
mfclose(SMFICTX *ctx)
{
    struct privdata *data = DATA;

    DEBUG_ENTER("mfclose", __LINE__);
    cleanup(ctx);
    if (data) {
	if (data->dir) free(data->dir);
	if (data->fp) fclose(data->fp);
	if (data->headerFP) fclose(data->headerFP);
	if (data->recipFP) fclose(data->recipFP);
	if (data->hostname) free(data->hostname);
	if (data->hostip) free(data->hostip);
	free(data);
    }
    smfi_setpriv(ctx, NULL);
    DEBUG_EXIT("mfclose", __LINE__, "SMFIS_CONTINUE");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: mfabort
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Called if current message is aborted.  Just cleans up.
***********************************************************************/
static sfsistat
mfabort(SMFICTX *ctx)
{
    return cleanup(ctx);
}

/**********************************************************************
*%FUNCTION: cleanup
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Cleans up temporary files.
***********************************************************************/
static sfsistat
cleanup(SMFICTX *ctx)
{
    sfsistat r = SMFIS_CONTINUE;
    struct privdata *data = DATA;
    pid_t n;

    DEBUG_ENTER("cleanup", __LINE__);
    if (!data) {
	DEBUG_EXIT("cleanup", __LINE__, "SMFIS_CONTINUE");
	return r;
    }

    if (data->fp && (fclose(data->fp) == EOF)) {
	DEBUG(syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__));
	r = SMFIS_TEMPFAIL;
    }
    data->fp = NULL;

    if (data->headerFP && (fclose(data->headerFP) == EOF)) {
	DEBUG(syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__));
	r = SMFIS_TEMPFAIL;
    }
    data->headerFP = NULL;

    if (data->recipFP && (fclose(data->recipFP) == EOF)) {
	DEBUG(syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__));
	r = SMFIS_TEMPFAIL;
    }
    data->recipFP = NULL;

    /* Remove temporary files unless "-d" command-line switch given */
    if (!DebugMode) {
	if (data->dir) {
	    n = fork();
	    if (n < 0) {
		DEBUG(syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__));
		r = SMFIS_TEMPFAIL;
	    } else if (n == 0) {
		/* In the child */
		closefiles();
		execl(RM, "rm", "-r", "-f", data->dir, NULL);
		_exit(EXIT_FAILURE);
	    } else {
		/* In the parent */
		if (waitpid(n, NULL, 0) < 0) {
		    DEBUG(syslog(LOG_ERR, "Failure in cleanup line %d: %m", __LINE__));
		    r = SMFIS_TEMPFAIL;
		}
	    }
	}
    }

    if (data->dir) {
	free(data->dir);
	data->dir = NULL;
    }

    DEBUG(syslog(LOG_INFO, "Exiting cleanup (line %d) ret=%s", __LINE__, (r == SMFIS_TEMPFAIL ? "SMFIS_TEMPFAIL" : "SMFIS_CONTINUE")));
    return r;
}

static struct smfiDesc filterDescriptor =
{
    "MIMEDefang-" VERSION,      /* Filter name */
    SMFI_VERSION,		/* Version code */

#if SMFI_VERSION == 1
    SMFIF_MODHDRS|SMFIF_MODBODY, /* We add a header and may alter body */
#else
    SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_MODBODY, /* We add a header and may alter body */
#endif

    mfconnect,			/* connection */
    NULL,			/* HELO */
    envfrom,			/* MAIL FROM: */
    rcptto,			/* RCPT TO: */
    header,			/* Called for each header */
    eoh,			/* Called at end of headers */
    body,			/* Called for each body chunk */
    eom,			/* Called at end of message */
    mfabort,			/* Called on abort */
    mfclose			/* Called on connection close */
};

/**********************************************************************
* %FUNCTION: main
* %ARGUMENTS:
*  argc, argv -- the usual suspects
* %RETURNS:
*  Whatever smfi_main returns
* %DESCRIPTION:
*  Main program
***********************************************************************/
int
main(int argc, char **argv)
{
    char c;
    int maxConn;

    /* Paranoia time */
    umask(077);
    /* Process command line options */
    while ((c = getopt(argc, argv, "n:p:d")) != -1) {
	switch (c) {
	case 'd':
	    DebugMode = 1;
	    break;

	case 'n':
	    if (optarg == NULL || sscanf(optarg, "%d", &maxConn) != 1) {
		fprintf(stderr, "Usage: mimedefang -n max_filters\n");
		exit(EXIT_FAILURE);
	    }
	    if (maxConn <= 0) {
		fprintf(stderr, "Usage: mimedefang -n max_filters (max_filters > 0)\n");
		exit(EXIT_FAILURE);
	    }
	    MaxRunningFilters = maxConn;
	    break;
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		(void) fprintf(stderr, "Illegal conn: %s\n",
			       optarg);
		exit(EXIT_FAILURE);
	    }
	    if (smfi_setconn(optarg) != MI_SUCCESS) {
		fprintf(stderr, "Could not open connection %s: %s",
			optarg, strerror(errno));
		exit(EXIT_FAILURE);
	    }
	    break;
	}
    }
    if (smfi_register(filterDescriptor) == MI_FAILURE) {
	fprintf(stderr, "smfi_register failed\n");
	exit(EXIT_FAILURE);
    }
    (void) closelog();
    openlog("mimedefang", LOG_PID, LOG_MAIL);
#ifdef ENABLE_DEBUGGING
    signal(SIGSEGV, handle_sig);
    signal(SIGBUS, handle_sig);
#endif
    return smfi_main();
}
